#include "sines.h"
#include <stdlib.h>
#include <iostream>
#include <math.h>
using namespace std;

//constructor
Sines::Sines() {
	m_play = false;
	m_inc = 0;
	for(int i = 0; i < MaxNumSines; i++) {
		m_phi[i] = 0.0;
	}
}

//destructor
Sines::~Sines() {
}

void Sines::initialize(bool play, float g_inc, int NumPts, float PointArray[][2]) {
	m_play = play;
	m_inc = g_inc;
	m_numPts = NumPts;
	
	for(int i = 0; i < NumPts; i++) {
		m_PointArray[i][0] = PointArray[i][0];
		m_PointArray[i][1] = PointArray[i][1];
	}
}

float Sines::setFreq(float xPos1, float yPos1, float xPos2, float yPos2, float xNow) {
	float slope = (yPos2 - yPos1) / (xPos2 - xPos1);
	//correct values too high or nan
	if(slope > 200) {
		slope = 200;
	} else if(isnan(slope)) {
		slope = 0;
	}
	float freq = yPos2 - (slope * (xPos2 - xNow));
	freq = 50 + 200 * pow(freq,0.5);
	//printf("%10.10f\n", freq);
	return freq;
}

void Sines::synthesize(SAMPLE * buff, int numFrames, int lineNumber) {
	//if playing, generate non zero samples
	if(m_play) {
		lineNumber *= 2;
		//fill
		for(int i = 0; i < numFrames; i++) {
			//if in range of a line, generate sound
			if(m_inc >= m_PointArray[lineNumber][0] && m_inc <= m_PointArray[lineNumber+1][0]) {
				//set target freq
				m_freq[lineNumber] = setFreq(m_PointArray[lineNumber][0], m_PointArray[lineNumber][1], 
									m_PointArray[lineNumber+1][0], m_PointArray[lineNumber+1][1], m_inc);		
				//generate sinewave
				m_phi[lineNumber] = m_phi[lineNumber] + (2 * MY_PI / (float)AUDIO_SRATE) * m_freq[lineNumber];
				m_phi[lineNumber] = fmod(m_phi[lineNumber], 2 * MY_PI);
				buff[i*NUMCHANNELS] = sin(m_phi[lineNumber]);
				//copy into second channel
				for(int j = 1; j < NUMCHANNELS; j++) {
					buff[i*NUMCHANNELS + j] = buff[i*NUMCHANNELS];
				}
			}
		}
		return;
	}
	//not playing, so silence
	for(int i = 0; i < numFrames; i++) {
		buff[i*NUMCHANNELS] = 0;
		//copy into second channel
		for(int j = 1; j < NUMCHANNELS; j++) {
			buff[i*NUMCHANNELS + j] = buff[i*NUMCHANNELS];
		}
	}
	return;
}
