//***************************************************************************
//xenakisSynth
//by Elliot Kermit-Canfield
//Autumn 2013
//***************************************************************************
#ifdef __MACOSX_CORE__
  #include <GLUT/glut.h>
#else
  #include <GL/gl.h>
  #include <GL/glu.h>
  #include <GL/glut.h>
#endif

#include <iostream>
#include <math.h>
#include <vector>
#include <cstdio>
#include <stdlib.h>

#include "graphics.h"
#include "globals.h"
#include "RtAudio.h"
#include "sines.h"

using namespace std;

// initial window size (in pixels)
GLsizei g_windowWidth = 800; 
GLsizei g_windowHeight = 600;
GLboolean g_fullscreen = false;
long g_last_width = g_windowWidth;
long g_last_height = g_windowHeight;

//drawing lines variables
float PointArray[MaxNumPts][2];
int NumPts = 0;

//for moving the line along the x-axis
float g_inc = 0;
float g_incSpeed = 1;

//instances of sines
Sines g_sines[MaxNumSines];

//audio varibles
SAMPLE g_freq = 440;
SAMPLE g_t = 0;
bool play = false;
//intermediate buffer
SAMPLE g_intermediate[FRAMESIZE];

GLint mouseCount = 0;

//-----------------------------------------------------------------------------
// Defines a point in a 3D space (coords x, y and z)
//-----------------------------------------------------------------------------
struct pt3d {
    pt3d( GLfloat _x, GLfloat _y, GLfloat _z ) : x(_x), y(_y), z(_z) {};
    
    float x;
    float y;
    float z;
};

// =======================
// = Function prototypes =
// =======================
void idleFunc( );
void displayFunc( );
void reshapeFunc( int width, int height );
void keyboardFunc( unsigned char, int, int );
void mouseFunc( int button, int state, int x, int y );
void specialFunc( int key, int x, int y );
void initialize( );
void changeLookAt( pt3d look_from, pt3d look_to, pt3d head_up );

void addNewPoint(float x, float y);
void removeFirstPoint();
void removeLastPoint();
void displayLines();
void drawGrid();
void drawScrollbar();

int sinewave(void * outputBuffer, void * inputBuffer, unsigned int numFrames, double streamTime, RtAudioStreamStatus status, void * data);
float setFreq(float xPos1, float yPos1, float xPos2, float yPos2, float xNow);
float setPhi(float xPos1, float yPos1, float xPos2, float yPos2, float xNow);

double rand2f(double low, double high);

// Camera global variables
pt3d g_look_from( 0, 0, 1 );
pt3d g_look_to( 0, 0, 0 );
pt3d g_head_up( 0, 1, 0 );

//----------------------------------------------------------------------------
// Name: rand2f();
//----------------------------------------------------------------------------
double rand2f(double low, double high){
    // swap
    if( low > high ) {
        double temp = low; low = high; high = temp;
    }
    // diff
    double diff = high - low;
    return low + diff * random() / (double)0x7fffffff;	
}


//-----------------------------------------------------------------------------
// Name: initialize( )
// Desc: sets initial OpenGL states
//       also initializes any application data
//-----------------------------------------------------------------------------
void initialize() {
    // set the GL clear color - use when the color buffer is cleared
    glClearColor( 0.0f, 0.0f, 0.0f, 1.0f );
        
    // set the shading model to 'smooth'
    glShadeModel( GL_SMOOTH );
    // enable depth
    glEnable( GL_DEPTH_TEST );
    // set the front faces of polygons
    glFrontFace( GL_CCW );
    // set fill mode
    glPolygonMode( GL_FRONT_AND_BACK, GL_FILL );

	//smooth lines and points
	glEnable(GL_POINT_SMOOTH);
	glEnable(GL_LINE_SMOOTH);
	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);

    // Enable transparency
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
      
    // seed random number generator
    srand( time(NULL) );
}

//-----------------------------------------------------------------------------
// Name: reshapeFunc( )
// Desc: called when window size changes
//-----------------------------------------------------------------------------
void reshapeFunc( int w, int h ) {
    // save the new window size
    g_windowWidth = (GLsizei)w; 
    g_windowHeight = (GLsizei)h;
    // map the view port to the client area
    glViewport( 0, 0, (GLsizei)w, (GLsizei)h );
    
    // set the matrix mode to project
    glMatrixMode( GL_PROJECTION );
    // load the identity matrix
    glLoadIdentity( );
    // create the viewing frustum
    gluPerspective( 90, (GLfloat)g_windowWidth / (GLfloat)g_windowHeight, 0.1, 300.0 );
    // set the matrix mode to modelview
    glMatrixMode( GL_MODELVIEW );
    // load the identity matrix
    glLoadIdentity( );
    // position the view point
    changeLookAt( g_look_from,g_look_to, g_head_up );

}

//-----------------------------------------------------------------------------
// Name: keyboardFunc( )
// Desc: key event
//-----------------------------------------------------------------------------
void keyboardFunc( unsigned char key, int x, int y )
{
    switch( key )
    {
    case 'Q':
    case 'q':
        exit(1);
        break;
	case 'd':
		for(int i = NumPts; i > 0; i--) {
			removeFirstPoint();
		}
		mouseCount = 0;
		break;
	case 'j':
		g_incSpeed += 0.1;
		break;
	case 'k':
		if(g_incSpeed > 0.1) {
			g_incSpeed -= 0.1;
		}
		break;
	case 'l':
		g_incSpeed = 1;
		break;
	case ' ':
		if(!play) {
			g_inc = 0;
			play = true;
		} else {
			play = false;
			g_inc = 0;
		}
		break;
	case 'f':
		if(!g_fullscreen) {
			g_last_width = g_windowWidth;
			g_last_height = g_windowHeight;
			g_fullscreen = true;
			glutFullScreen();
		} else {
			g_fullscreen = false;
			glutReshapeWindow( g_last_width, g_last_height );
		}
		break;
    }

    glutPostRedisplay( );
}

//-----------------------------------------------------------------------------
// Name: mouseFunc( )
// Desc: handles mouse stuff
//-----------------------------------------------------------------------------
void mouseFunc( int button, int state, int x, int y ) {
    if( button == GLUT_LEFT_BUTTON ) {
		float xPos = 10 * ((float)x/((float)g_windowWidth - 1));
		float yPos = 10 * ((float)y/((float)g_windowHeight - 1)) -10;
		yPos = - yPos;
		addNewPoint(xPos, yPos);
		mouseCount = mouseCount + 1;
		if(mouseCount % 2 == 0) {
			//correct where points are if order is wrong
			for(int i = 0; i < mouseCount; i+=2) {
				if(PointArray[i][0] > PointArray[i+1][0]) {
					float tempx = PointArray[i][0];
					float tempy = PointArray[i][1];
					PointArray[i][0] = PointArray[i+1][0];
					PointArray[i][1] = PointArray[i+1][1];
					PointArray[i+1][0] = tempx;
					PointArray[i+1][1] = tempy;
				}
			}
		}
		//cerr << "count: " << mouseCount << endl;
		//cerr << "xpos:" << xPos << " ypos:" << yPos << endl;
    }
    else if ( button == GLUT_RIGHT_BUTTON ) {
        // when right mouse button down, move right
        if( state == GLUT_DOWN ) {
        }
        else {
        }
    }
    else {
    }

    glutPostRedisplay( );
}

//-------------------------------------------------------------------------------
//drawing lines functions
//-------------------------------------------------------------------------------
void removeFirstPoint() {
	int i;
	if(NumPts > 0) {
		//remove first point and slide the rest down
		NumPts--;
		for(i = 0; i < NumPts; i++) {
			PointArray[i][0] = PointArray[i+1][0];
			PointArray[i][1] = PointArray[i+1][1];
		}
	}
}

void removeLastPoint() {
	if(NumPts > 0) {
		NumPts--;
	}
}

void addNewPoint(float x, float y) {
	if(NumPts >= MaxNumPts) {
		removeFirstPoint();
	}	
	PointArray[NumPts][0] = x;
	PointArray[NumPts][1] = y;
	NumPts++;
}

void displayLines() {
	int i;
	glColor3f(1.0f, 1.0f, 1.0f);
	if(NumPts > 1) {
		glBegin(GL_LINES);
		for(i = 0; i < NumPts; i++) {
			if(play) {
				glColor3f(rand2f(0,1), rand2f(0,1), rand2f(0,1));
			}
			glVertex2f(PointArray[i][0], PointArray[i][1]);
		}
		glEnd();
	}
	glColor3f(1.0f, 1.0f, 1.0f);
	glBegin(GL_POINTS);
	for(i = 0; i < NumPts; i++) {
		glVertex2f(PointArray[i][0], PointArray[i][1]);
	}
	glEnd();
}

void drawGrid() {
    // set the view point
    changeLookAt( g_look_from, g_look_to, g_head_up );  
	glColor3f(.5,.5,.5);
	glBegin(GL_LINES);
	for(int i=0; i<=10; i++){
		glVertex2f(i, 0);
		glVertex2f(i, 10);	
		glVertex2f(0, i);
		glVertex2f(10, i);
	}
	glEnd();
}

void drawScrollbar() {
	glColor3f(1.0f, 0, 0);
	glBegin(GL_LINES);
	glVertex2f(0+g_inc, 10);
	glVertex2f(0+g_inc, 0);
	g_inc = g_inc + (0.001 * g_incSpeed);
	if(g_inc >= 10) {
		g_inc = 0;
	}
	glEnd();
}


//-----------------------------------------------------------------------------
// Name: specialFunc( )
// Desc: handles special function keys
//-----------------------------------------------------------------------------
void specialFunc( int key, int x, int y )
{
    if( key == GLUT_KEY_LEFT) {
        std::cerr << "Left arrow" << std::endl;
    }
    if( key == GLUT_KEY_RIGHT) {
        std::cerr << "Right arrow" << std::endl;
    }
    if( key == GLUT_KEY_DOWN) {
        std::cerr << "Down arrow" << std::endl;
    }
    if( key == GLUT_KEY_UP) {
        std::cerr << "Up arrow" << std::endl;
    }
    if( key == GLUT_KEY_PAGE_UP) {
        std::cerr << "PageUp arrow" << std::endl;
    }
    if( key == GLUT_KEY_PAGE_DOWN) {
        std::cerr << "PageDown arrow" << std::endl;
    }

    glutPostRedisplay( );
}


//-----------------------------------------------------------------------------
// Name: idleFunc( )
// Desc: callback from GLUT
//-----------------------------------------------------------------------------
void idleFunc( )
{
    // render the scene
    glutPostRedisplay( );
}


//-----------------------------------------------------------------------------
// Name: displayFunc( )
// Desc: callback function invoked to draw the client area
//-----------------------------------------------------------------------------
void displayFunc( )
{
    // clear the color and depth buffers
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    // set line width
    glLineWidth( 1.0f );

    //set 2d mode
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0f, 10.0f, 0.0f, 10.0f, 0.0f, 30.0);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

//draw a grid    
    glPushMatrix();
		drawGrid();
    glPopMatrix();

//draw the pitch contours
	glPushMatrix();
		glPointSize(5);
		glLineWidth(3);
		displayLines();
	glPopMatrix();
	
//draw the scrolling bar
	if(play) {
		glPushMatrix();
			drawScrollbar();
		glPopMatrix();
	}
  
    // flush!
    glFlush( );
    // swap the double buffer
    glutSwapBuffers( );
}

//-----------------------------------------------------------------------------
// name: changeLookAt()
// desc: changes the "camera"
//-----------------------------------------------------------------------------
void changeLookAt( pt3d look_from, pt3d look_to, pt3d head_up )
{
    gluLookAt(  look_from.x, look_from.y, look_from.z, 
                look_to.x, look_to.y, look_to.z, 
                head_up.x, head_up.y, head_up.z );
}

//---------------------------------------------------------------------------
// name: xenakis_about()
//---------------------------------------------------------------------------
void xenakis_about() {
	fprintf( stderr, "------------------------------------------------------------------------------\n");
	fprintf( stderr, "><<      ><< ><<<<<<<< ><<<     ><<       ><        ><<   ><<   ><<   ><< <<\n");
	fprintf( stderr, " ><<   ><<   ><<       >< ><<   ><<      >< <<      ><<  ><<    ><< ><<    ><<\n");
	fprintf( stderr, "  ><< ><<    ><<       ><< ><<  ><<     ><  ><<     ><< ><<     ><<  ><<\n");
	fprintf( stderr, "    ><<      ><<<<<<   ><<  ><< ><<    ><<   ><<    >< ><       ><<    ><<\n");   
	fprintf( stderr, "  ><< ><<    ><<       ><<   >< ><<   ><<<<<< ><<   ><<  ><<    ><<       ><<\n");
	fprintf( stderr, " ><<   ><<   ><<       ><<    >< <<  ><<       ><<  ><<   ><<   ><< ><<    ><<\n");
	fprintf( stderr, "><<      ><< ><<<<<<<< ><<      ><< ><<         ><< ><<     ><< ><<   ><< <<\n\n");
	fprintf( stderr, "        ><< <<    ><<      ><<  ><<<     ><<  ><<< ><<<<<<  ><<     ><<\n");
	fprintf( stderr, "      ><<    ><<   ><<    ><<   >< ><<   ><<       ><<      ><<     ><<\n");
	fprintf( stderr, "       ><<          ><< ><<     ><< ><<  ><<       ><<      ><<     ><<\n");
	fprintf( stderr, "         ><<          ><<       ><<  ><< ><<       ><<      ><<<<<< ><<\n");
	fprintf( stderr, "            ><<       ><<       ><<   >< ><<       ><<      ><<     ><<\n");
	fprintf( stderr, "      ><<    ><<      ><<       ><<    >< <<       ><<      ><<     ><<\n");
	fprintf( stderr, "        ><< <<        ><<       ><<      ><<       ><<      ><<     ><<\n");
	fprintf( stderr, "------------------------------------------------------------------------------\n");
    fprintf( stderr, "      | by Elliot Kermit-Canfield\n" );
    fprintf( stderr, "      | Stanford University | CCRMA\n" );
    fprintf( stderr, "      | http://ccrma.stanford.edu/~kermit/\n" );
    fprintf( stderr, "      | version: 0.0.1\n\n" );
}

//-----------------------------------------------------------------------------
// name: xenakis_keys()
//-----------------------------------------------------------------------------
void xenakis_keys() {
    xenakis_line();
    fprintf( stderr, "[xenkais synth]: run-time control\n" );
    xenakis_line();
    fprintf( stderr, "  print messages\n" );
    fprintf( stderr, "  and some more\n" );
}

//-----------------------------------------------------------------------------
// name: xenakis_help()
//-----------------------------------------------------------------------------
void xenakis_help() {
    xenakis_endline();
    xenakis_keys();
    xenakis_endline();
}

//-----------------------------------------------------------------------------
// name: xenakis_usage()
//-----------------------------------------------------------------------------
void xenakis_usage() {
    xenakis_line();
    fprintf( stderr, "[xenakis]: command line arguments\n" );
    xenakis_line();
}

//-----------------------------------------------------------------------------
// name: xenakis_line()
// desc: ...
//-----------------------------------------------------------------------------
void xenakis_line() {
    fprintf( stderr, "---------------------------------------------------------\n" );
}

//-----------------------------------------------------------------------------
// name: xenakis_endline()
// desc: ...
//-----------------------------------------------------------------------------
void xenakis_endline() {
    fprintf( stderr, "\n" );
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
//-----------------------------------------------------------------------------
// name: sinewave()
// desc: sinewave audio callback
//-----------------------------------------------------------------------------
int sinewave(void * outputBuffer, void * inputBuffer, unsigned int numFrames, double streamTime, RtAudioStreamStatus status, void * data) {
	//cast
	SAMPLE * output = (SAMPLE *)outputBuffer;
	
	//zero out
	for(int j = 0; j < numFrames*NUMCHANNELS; j++) {
		output[j] = 0;
		g_intermediate[j] = 0;
	}
	//synthesize
	for(int i = 0; i < NumPts/2; i++) {
		for(int j = 0; j < numFrames*NUMCHANNELS; j++) {
			g_intermediate[j] = 0;
		}
		g_sines[i].initialize(play, g_inc, NumPts, PointArray);
		g_sines[i].synthesize(g_intermediate, numFrames, i);
		
		//accumulate into outputbuffer
		for(int j = 0; j < numFrames*NUMCHANNELS; j++) {
			output[j] += g_intermediate[j];
		}
	}
	return 0;
}


//---------------------------------------------------------------------------
//name: main()
//desc: main enty point
//---------------------------------------------------------------------------
int main(int argc, char ** argv) {
	//start RtAudio
	RtAudio adac;
	//variables
	unsigned int bufferBytes = 0;
	//buffer framesize
	unsigned int bufferFrames = 512;

	//check for audio devices
	if(adac.getDeviceCount() < 1) {
		cout << "no audio devices" << endl;
		exit(1);
	}
	
	//let RtAudio print messages to stdout
	adac.showWarnings(true);
	
	//set input/output parameters
	RtAudio::StreamParameters iParams, oParams;
	iParams.deviceId = adac.getDefaultInputDevice();
	iParams.nChannels = NUMCHANNELS;
	iParams.firstChannel = 0;
	oParams.deviceId = adac.getDefaultOutputDevice();
	oParams.nChannels = NUMCHANNELS;
	oParams.firstChannel = 0;
	
    //create stream options
	RtAudio::StreamOptions options;

	//open stream
    try{
        //open stream
        adac.openStream(&oParams, &iParams, MY_FORMAT, AUDIO_SRATE, &bufferFrames, &sinewave, (void *)&bufferBytes, &options);
    }
    catch(RtError& error) {
        cout << error.getMessage() << endl;
        exit(1);
    }

    //compute wave
    bufferBytes = bufferFrames * NUMCHANNELS * sizeof(SAMPLE);
    
    //test RtAudio functionality for reporting latency
    cout << "stream latency: " << adac.getStreamLatency() << "frames" << endl;
    
    try{
        //start stream
        adac.startStream();
    }
    catch(RtError& error) {
        cout << error.getMessage() << endl;
        goto cleanup;
	}


	
    // initialize GLUT
    glutInit( &argc, argv );
    // double buffer, use rgb color, enable depth buffer
    glutInitDisplayMode( GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH );
    // initialize the window size
    glutInitWindowSize( g_windowWidth, g_windowHeight );
    // set the window postion
    glutInitWindowPosition( 50, 50 );
    // create the window
    glutCreateWindow( "XENAKIS SYNTH" );
    //glutEnterGameMode();

    // set the idle function - called when idle
    glutIdleFunc( idleFunc );
    // set the display function - called when redrawing
    glutDisplayFunc( displayFunc );
    // set the reshape function - called when client area changes
    glutReshapeFunc( reshapeFunc );
    // set the keyboard function - called on keyboard events
    glutKeyboardFunc( keyboardFunc );
    // set the mouse function - called on mouse stuff
    glutMouseFunc( mouseFunc );
    // set the special function - called on special keys events (fn, arrows, pgDown, etc)
    glutSpecialFunc( specialFunc );

    // do our own initialization
    initialize();

	//about the program
	xenakis_about();

    // let GLUT handle the current thread from here
    glutMainLoop();

	cleanup:
	//close if open
	if(adac.isStreamOpen() )
		adac.closeStream();
		
	return 0;
}
